from __future__ import annotations

from typing import Any

import lightgbm as lgb
import numpy as np
import pandas as pd
import pytest
from factrainer.core import CvModelContainer, SingleModelContainer
from factrainer.lightgbm import LgbDataset, LgbModelConfig, LgbTrainConfig
from numpy import typing as npt
from sklearn.metrics import accuracy_score, r2_score
from sklearn.model_selection import KFold, train_test_split


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(target, y_pred)

    assert (metric > 0.8) and (metric < 0.85)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_model_parallel(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold, n_jobs_train=4, n_jobs_predict=4)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = r2_score(target, y_pred)

    assert (metric > 0.8) and (metric < 0.85)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_cv_pandas(titanic_data: tuple[pd.DataFrame, pd.Series[int]]) -> None:
    features, target = titanic_data
    dataset = LgbDataset(dataset=lgb.Dataset(features, label=target))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "binary"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    k_fold = KFold(n_splits=4, shuffle=True, random_state=1)
    model = CvModelContainer(config, k_fold, n_jobs_train=4, n_jobs_predict=4)
    model.train(dataset)
    y_pred = model.predict(dataset)
    metric = accuracy_score(target, y_pred > 0.5)

    assert (metric > 0.8) and (metric < 0.85)


@pytest.mark.flaky(reruns=3, reruns_delay=5, only_rerun=["HTTPError"])
def test_single_model(
    california_housing_data: tuple[
        npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
    ],
) -> None:
    features, target = california_housing_data
    train_X, test_X, train_y, test_y = train_test_split(
        features,
        target,
        test_size=0.2,
        random_state=1,
    )
    train_dataset = LgbDataset(dataset=lgb.Dataset(train_X, train_y))
    val_dataset = LgbDataset(dataset=lgb.Dataset(test_X, test_y))
    test_dataset = LgbDataset(dataset=lgb.Dataset(test_X, test_y))
    config = LgbModelConfig.create(
        train_config=LgbTrainConfig(
            params={"objective": "regression"},
            callbacks=[lgb.early_stopping(100, verbose=False)],
        ),
    )
    model = SingleModelContainer(config)
    model.train(train_dataset, val_dataset)
    y_pred = model.predict(test_dataset)
    metric = r2_score(test_y, y_pred)

    assert (metric > 0.8) and (metric < 0.85)



from __future__ import annotations

from typing import Any

import numpy as np
import pandas as pd
import pytest
from numpy import typing as npt
from sklearn.datasets import fetch_california_housing, fetch_openml


@pytest.fixture(scope="package")
def california_housing_data() -> tuple[
    npt.NDArray[np.number[Any]], npt.NDArray[np.number[Any]]
]:
    data = fetch_california_housing()
    return data["data"], data["target"]  # type: ignore


@pytest.fixture(scope="package")
def titanic_data() -> tuple[pd.DataFrame, pd.Series[int]]:
    titanic = fetch_openml(name="titanic", version=1, as_frame=True)
    features = titanic["data"].drop(  # type: ignore
        columns=["name", "ticket", "cabin", "boat", "body", "home.dest"]
    )
    target = titanic["target"].astype("int")  # type: ignore
    return features, target



from typing import Any
from unittest.mock import MagicMock, patch

import lightgbm as lgb
import numpy as np
from factrainer.lightgbm.dataset.slicer import (
    LgbDatasetSlicer,
    LgbDataSlicer,
    LgbGroupSlicer,
    LgbInitScoreSlicer,
    LgbLabelSlicer,
    LgbPositionSlicer,
    LgbWeightSlicer,
)
from lightgbm.basic import (
    _LGBM_CategoricalFeatureConfiguration,
    _LGBM_FeatureNameConfiguration,
    _LGBM_GroupType,
    _LGBM_InitScoreType,
    _LGBM_LabelType,
    _LGBM_PositionType,
    _LGBM_TrainDataType,
    _LGBM_WeightType,
)
from numpy.testing import assert_array_equal


@patch("factrainer.lightgbm.dataset.slicer.LgbDataSlicer", spec=LgbDataSlicer)
def test_dataset_slicer(data_slicer: MagicMock) -> None:
    data = MagicMock(spec=_LGBM_TrainDataType)
    dataset = lgb.Dataset(data)
    expected = lgb.Dataset(
        data=data_slicer.return_value.slice.return_value,
    )
    sut = LgbDatasetSlicer()

    actual = sut.slice(dataset, [2, 0])

    assert actual.data == expected.data
    assert actual.label is None
    assert expected.reference is None
    assert actual.weight is None
    assert actual.group is None
    assert actual.init_score is None
    assert actual.feature_name == "auto"
    assert actual.categorical_feature == "auto"
    assert actual.free_raw_data is True
    data_slicer.return_value.slice.assert_called_once_with(data, [2, 0])


@patch("factrainer.lightgbm.dataset.slicer.LgbPositionSlicer", spec=LgbPositionSlicer)
@patch("factrainer.lightgbm.dataset.slicer.LgbInitScoreSlicer", spec=LgbInitScoreSlicer)
@patch("factrainer.lightgbm.dataset.slicer.LgbGroupSlicer", spec=LgbGroupSlicer)
@patch("factrainer.lightgbm.dataset.slicer.LgbWeightSlicer", spec=LgbWeightSlicer)
@patch("factrainer.lightgbm.dataset.slicer.LgbLabelSlicer", spec=LgbLabelSlicer)
@patch("factrainer.lightgbm.dataset.slicer.LgbDataSlicer", spec=LgbDataSlicer)
def test_dataset_slicer_with_params(
    data_slicer: MagicMock,
    label_slicer: MagicMock,
    weight_slicer: MagicMock,
    group_slicer: MagicMock,
    init_score_slicer: MagicMock,
    position_slicer: MagicMock,
) -> None:
    dataset = lgb.Dataset(
        data=MagicMock(spec=_LGBM_TrainDataType),
        label=MagicMock(spec=_LGBM_LabelType),
        weight=MagicMock(spec=_LGBM_WeightType),
        group=MagicMock(spec=_LGBM_GroupType),
        init_score=MagicMock(spec=_LGBM_InitScoreType),
        feature_name=MagicMock(spec=_LGBM_FeatureNameConfiguration),
        categorical_feature=MagicMock(spec=_LGBM_CategoricalFeatureConfiguration),
        params=MagicMock(spec=dict[str, Any]),
        free_raw_data=MagicMock(spec=bool),
        position=MagicMock(spec=_LGBM_PositionType),
    )
    expected = lgb.Dataset(
        data=data_slicer.return_value.slice.return_value,
        label=label_slicer.return_value.slice.return_value,
        reference=None,
        weight=weight_slicer.return_value.slice.return_value,
        group=group_slicer.return_value.slice.return_value,
        init_score=init_score_slicer.return_value.slice.return_value,
        feature_name=dataset.feature_name,
        categorical_feature=dataset.categorical_feature,
        params=dataset.params,
        free_raw_data=dataset.free_raw_data,
        position=position_slicer.return_value.slice.return_value,
    )
    sut = LgbDatasetSlicer()

    actual = sut.slice(dataset, [2, 0])

    assert actual.data == expected.data
    assert actual.label == expected.label
    assert expected.reference is None
    assert actual.weight == expected.weight
    assert actual.group == expected.group
    assert actual.init_score == expected.init_score
    assert actual.feature_name == expected.feature_name
    assert actual.categorical_feature == expected.categorical_feature
    assert actual.free_raw_data == expected.free_raw_data
    assert actual.position == expected.position
    data_slicer.return_value.slice.assert_called_once_with(dataset.data, [2, 0])
    label_slicer.return_value.slice.assert_called_once_with(dataset.label, [2, 0])
    weight_slicer.return_value.slice.assert_called_once_with(dataset.weight, [2, 0])
    group_slicer.return_value.slice.assert_called_once_with(dataset.group, [2, 0])
    init_score_slicer.return_value.slice.assert_called_once_with(
        dataset.init_score, [2, 0]
    )
    position_slicer.return_value.slice.assert_called_once_with(dataset.position, [2, 0])


def test_dataset_slicer_with_reference() -> None:
    dataset = lgb.Dataset(np.array([[100, 200], [300, 400]]))
    reference = lgb.Dataset(np.array([[1, 2], [3, 4]]))
    expected = lgb.Dataset(np.array([[1, 2], [3, 4]]))
    sut = LgbDatasetSlicer(reference=reference)

    actual = sut.slice(dataset, [1])

    if actual.reference is None:
        raise TypeError
    if not isinstance(actual.reference.data, np.ndarray):
        raise TypeError
    if not isinstance(expected.data, np.ndarray):
        raise TypeError
    assert_array_equal(actual.reference.data, expected.data)



import lightgbm as lgb
import numpy as np
import pandas as pd
from factrainer.lightgbm import LgbDataset


class TestLgbDatasetGetitem:
    def test_single_row(self) -> None:
        sut = LgbDataset(dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6]])))
        expected = LgbDataset(dataset=lgb.Dataset(data=np.array([[1, 2, 3]])))
        actual = sut[0]
        assert actual == expected

    def test_rows(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(data=np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]))
        )
        expected = LgbDataset(
            dataset=lgb.Dataset(data=np.array([[7, 8, 9], [1, 2, 3]]))
        )
        actual = sut[[2, 0]]
        assert actual == expected

    def test_pandas(self) -> None:
        sut = LgbDataset(
            dataset=lgb.Dataset(
                data=pd.DataFrame(
                    {
                        "a": [1, 2, 3],
                        "b": ["4", pd.NA, "6"],
                    }
                )
            )
        )
        expected = LgbDataset(
            dataset=lgb.Dataset(
                data=pd.DataFrame(
                    {
                        "a": [2, 1],
                        "b": [pd.NA, "4"],
                    },
                    index=[1, 0],
                )
            )
        )

        actual = sut[[1, 0]]

        assert actual == expected
